import { getGuildApplicationCommands } from "discord-api-methods";
import esbuild from "esbuild";
import fs from "node:fs";
import path from "node:path";
export async function getEdgeEntrypoint(filepath) {
    if (filepath != null) {
        return path.join(process.cwd(), filepath);
    }
    const presets = ["/src/server.ts", "/src/server.js", "/src/server.mjs"];
    for (const preset of presets) {
        if (fs.existsSync(path.join(process.cwd(), preset))) {
            return path.join(process.cwd(), preset);
        }
    }
    throw new Error(`Could not find server entrypoint. Please create a server file named one of the following: ${presets
        .map((name) => `\`${name}\``)
        .join(", ")}`);
}
export async function getApplicationEntrypoint() {
    try {
        const json = (await import(path.join(process.cwd(), "package.json"), {
            assert: { type: "json" },
        }));
        const appPath = json?.default?.main;
        // Prevent ESM Caching
        const cacheBuster = `?update=${Date.now()}`;
        let entrypoint = path.join(process.cwd(), appPath + cacheBuster);
        if (entrypoint.endsWith(".ts")) {
            const build = await esbuild.build({
                entryPoints: [entrypoint],
                bundle: true,
                format: "esm",
                outdir: path.join(process.cwd(), ".ikit"),
            });
            entrypoint = path.join(process.cwd(), ".ikit", appPath.replace("src/", "").replace(".ts", ".js") + cacheBuster);
        }
        const app = (await import(entrypoint));
        return app?.default;
    }
    catch (error) {
        console.error("There was an error reading your application file:\n");
        console.error(error);
        process.exit(1);
    }
}
function getChangeSet(application, commandList) {
    const comparison = {
        changed: false,
        new: new Set(),
        updated: new Set(),
        deleted: new Set(),
        unchanged: new Set(),
    };
    // Compare all existing commands against registered ones
    for (const command of application.commands) {
        // If the command already exists, we check to see if it's changed or not
        if (commandList.has(command.name)) {
            const schema = commandList.get(command.name);
            if (schema == null) {
                continue;
            }
            if (command.equals(schema)) {
                comparison.unchanged.add(command.serialize());
            }
            else {
                comparison.updated.add(command.serialize());
                comparison.changed = true;
            }
            commandList.delete(command.name);
        }
        // If the command does not exist, we add it
        else {
            comparison.new.add(command.serialize());
            comparison.changed = true;
        }
    }
    // Any command left in the Discord list no longer exists in the code; eliminate them
    comparison.deleted = new Set(Array.from(commandList.values()));
    if (comparison.deleted.size > 0) {
        comparison.changed = true;
    }
    return comparison;
}
export async function getGuildApplicationCommandChanges(application, guildId, cachedCommands) {
    const commandList = new Map();
    if (cachedCommands == null) {
        const response = await getGuildApplicationCommands(application.id, guildId);
        response.forEach((cmd) => commandList.set(cmd.name, cmd));
    }
    else {
        cachedCommands.forEach((cmd) => commandList.set(cmd.name, cmd));
    }
    return getChangeSet(application, commandList);
}
//# sourceMappingURL=scripts.js.map